import logging

LOG_LEVELS = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


class PSLBase:
    def __init__(self, log_level=None, *args, **kwargs):
        if log_level is not None:
            log_level = LOG_LEVELS.get(log_level) or LOG_LEVELS[0]
            logging.basicConfig(level=log_level)

        self.logger = logging.getLogger('psl')



# Main library, used by everything
from psl_dns.querier import PSL
from psl_dns import exceptions

# Special-purpose classes, used by commands
from psl_dns.checker import Checker
from psl_dns.parser import Parser



import dns.rdatatype
import dns.resolver

from psl_dns import exceptions
from psl_dns.base import PSLBase
from psl_dns.providers import DefaultProvider


class PSL(PSLBase):
    def __init__(self, resolver=None, zone=DefaultProvider.ZONE, timeout=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        use_system_nameservers = resolver is None
        self.resolver = dns.resolver.Resolver(configure=use_system_nameservers)
        if not use_system_nameservers:
            self.resolver.nameservers = [resolver]
        self.timeout = timeout
        self.zone = zone.rstrip('.') + '.'

    def get_checksum(self):
        answer = self.query('', dns.rdatatype.TXT)
        if not len(answer):
            return None

        return answer[0].to_text().strip('"').split(' ')[1]

    def _get_public_suffix_raw(self, domain):
        # Retrieve RRset
        try:
            answer = self.query(domain, dns.rdatatype.PTR)
        except dns.resolver.NoAnswer:
            msg = 'Domain {} is affected by an unsupported rule'.format(domain)
            raise exceptions.UnsupportedRule(msg)

        return answer[0].to_text()

    def get_public_suffix(self, domain):
        if domain[0] == '.':
            raise ValueError('Invalid domain name')

        # Get public suffix and normalize
        public_suffix = self._get_public_suffix_raw(domain)
        if domain[-1] != '.':
            public_suffix = public_suffix[:-1]

        # Replace wildcard labels with the corresponding labels from the domain (punycode)
        punycode_domain = domain.encode('idna').decode('ascii')
        domain_labels = punycode_domain.split('.')
        domain_labels.reverse()
        public_labels = public_suffix.split('.')
        public_labels.reverse()

        for i, domain_label, public_label in zip(range(len(public_labels)), domain_labels, public_labels):
            domain_label = domain_label.lower()
            if public_label == '*':
                public_labels[i] = domain_label
            elif public_label != domain_label.encode('idna').decode('ascii'):
                raise ValueError(
                    'Public suffix label {} inconsistent with domain label {}'.format(public_label, domain_label))

        public_labels.reverse()
        public_suffix = '.'.join(public_labels)

        # Return in initial encoding
        punycode = (domain == punycode_domain)
        return public_suffix if punycode else public_suffix.encode('ascii').decode('idna')

    def get_rules(self, domain):
        # The public suffix rule itself is always a rule
        try:
            rules = [self._get_public_suffix_raw(domain).rstrip('.')]
        except exceptions.UnsupportedRule:
            rules = []

        # For wildcard exceptions and unsupported inline wildcards, additional rules
        # are given as TXT records.
        try:
            rules.extend([rr.to_text()[1:-1] for rr in self.query(domain, dns.rdatatype.TXT)])
        except dns.resolver.NoAnswer:
            pass

        return {str(rule.encode('utf-8'), 'idna') for rule in rules}

    def is_public_suffix(self, domain, public_suffix=None):
        public_suffix = public_suffix or self.get_public_suffix(domain)
        return domain.count('.') == public_suffix.count('.')

    def query(self, domain, rdatatype):
        # Normalize, then construct QNAME, and retrieve response
        qname = '.'.join([domain.lower().rstrip('.'), self.zone]).lstrip('.')
        qname = dns.name.from_text(qname)
        self.logger.info('Querying for {} {}'.format(qname, dns.rdatatype.to_text(rdatatype)))
        try:
            answer = self.resolver.query(qname, rdatatype, lifetime=self.timeout)
        except dns.resolver.NXDOMAIN:
            message = 'Cannot find {} {} record. (Resolver claims NXDOMAIN. Are you using a non-compliant resolver?)'
            message = message.format(qname, dns.rdatatype.to_text(rdatatype))
            raise exceptions.ResolverError(message)
        return answer



from unittest import mock, TestCase

import dns

import psl_dns


class TestBase(TestCase):
    def setUp(self):
        mock_resolver_class = mock.patch('dns.resolver.Resolver', autospec=True).start()
        self.mock_resolver = mock_resolver_class.return_value
        self.mock_resolver.query.return_value = [mock.create_autospec(dns.rdata.Rdata).return_value]
        self.addCleanup(mock.patch.stopall)

        self.psl = psl_dns.PSL()

    def get_mock_context_manager(self, value):
        if hasattr(value, '__call__'):
            return mock.patch.object(self.mock_resolver, 'query', side_effect=value)

        def _query_side_effect(_domain, rdatatype, **_):
            if isinstance(value, dict):
                records = value[rdatatype]
            else:
                records = value if isinstance(value, list) else [value]

            answers = []
            for record in records:
                answers.append(mock.create_autospec(dns.rdata.Rdata).return_value)
                answers[-1].to_text.return_value = record
            return answers

        return mock.patch.object(self.mock_resolver, 'query', side_effect=_query_side_effect)


class TestMisc(TestBase):
    def test_get_checksum(self):
        checksum = 'd205f587d61c6bbf05bec818776da1dd030ce68f2e8912fea732158b9a33cc54'
        with self.get_mock_context_manager('"1556058819 {}"'.format(checksum)):
            self.assertEqual(self.psl.get_checksum(), checksum)

    def test_nxdomain(self):
        with self.get_mock_context_manager(dns.resolver.NXDOMAIN):
            with self.assertRaises(psl_dns.exceptions.ResolverError):
                self.psl.query('iana.org', dns.rdatatype.PTR)

    def test_get_rules(self):
        mock_records = {dns.rdatatype.PTR: ['*.'], dns.rdatatype.TXT: ['"*.ck"', '"!www.ck"']}
        with self.get_mock_context_manager(mock_records):
            self.assertEqual(self.psl.get_rules('www.ck.'), {'!www.ck', '*.ck', '*'})
            self.mock_resolver.query.assert_called()


class TestGetPublicSuffix(TestBase):
    def assertGetPublicSuffix(self, domain, suffix):
        mock_answer = (suffix.rstrip('.') + '.').encode('idna').decode('ascii')
        with self.get_mock_context_manager(mock_answer):
            self.assertEqual(self.psl.get_public_suffix(domain), suffix)
            self.mock_resolver.query.assert_called()

    def test_ascii(self):
        self.assertGetPublicSuffix('io', 'io')
        self.assertGetPublicSuffix('dedyn.io', 'dedyn.io')
        self.assertGetPublicSuffix('desec.io', 'io')
        self.assertGetPublicSuffix('desec.io.', 'io.')

        # Also test lettercase normalization
        self.assertGetPublicSuffix('IO', 'io')
        self.assertGetPublicSuffix('s3.AmazonAWS.com', 's3.amazonaws.com')

    def test_punycode(self):
        self.assertGetPublicSuffix('www.xn--tsting-wxa.de.co.uk', 'co.uk')

    def test_unicode(self):
        self.assertGetPublicSuffix('www.tösting.co.uk', 'co.uk')
        self.assertGetPublicSuffix('www.xn--55qx5d.cn.', 'xn--55qx5d.cn.')
        self.assertGetPublicSuffix('www.公司.cn.', '公司.cn.')
        self.assertGetPublicSuffix('公司.cn', '公司.cn')

    def test_unsupported_rule_exception(self):
        with self.get_mock_context_manager(psl_dns.exceptions.UnsupportedRule):
            with self.assertRaises(psl_dns.exceptions.UnsupportedRule):
                self.psl.get_public_suffix('unsupported.wildcard.test')

    def test_invalid_domain(self):
        with self.assertRaises(ValueError):
            self.psl.get_public_suffix('.')

        with self.assertRaises(ValueError):
            self.psl.get_public_suffix('.desec.io')

    def test_inconsistent_labels(self):
        with self.get_mock_context_manager('dedyn.io.'):
            with self.assertRaises(ValueError):
                self.psl.get_public_suffix('www.desec.io')


class TestIsPublicSuffix(TestBase):
    def assertIsPublicSuffix(self, domain, suffix, value):
        mock_answer = (suffix.rstrip('.') + '.').encode('idna').decode('ascii')
        with self.get_mock_context_manager(mock_answer):
            self.assertEqual(self.psl.is_public_suffix(domain), value)
            self.mock_resolver.query.assert_called()

    def test_ascii(self):
        self.assertIsPublicSuffix('io', 'io', True)
        self.assertIsPublicSuffix('dedyn.io', 'dedyn.io', True)
        self.assertIsPublicSuffix('desec.io', 'io', False)
        self.assertIsPublicSuffix('desec.io.', 'io', False)

    def test_punycode(self):
        self.assertIsPublicSuffix('www.xn--tsting-wxa.de.co.uk', 'co.uk', False)

    def test_unicode(self):
        self.assertIsPublicSuffix('www.tösting.co.uk', 'co.uk', False)
        self.assertIsPublicSuffix('www.公司.cn.', '公司.cn', False)
        self.assertIsPublicSuffix('公司.cn', '公司.cn', True)

    def test_given_suffix(self):
        self.assertTrue(self.psl.is_public_suffix('something.ck', 'something.ck'))
        self.assertFalse(self.psl.is_public_suffix('something.ck', 'something.ck.'))
